@@ -1,6 +1,6 @@
 /* findcmd.h - functions from findcmd.c. */
 
-/* Copyright (C) 1997-2009 Free Software Foundation, Inc.
+/* Copyright (C) 1997-2012 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -29,7 +29,7 @@ extern int is_directory __P((const char *));
 extern int executable_or_directory __P((const char *));
 extern char *find_user_command __P((const char *));
 extern char *find_path_file __P((const char *));
-extern char *search_for_command __P((const char *));
+extern char *search_for_command __P((const char *, int));
 extern char *user_command_matches __P((const char *, int, int));
 
 #endif /* _FINDCMD_H_ */
