@@ -1,24 +1,24 @@
+/*
+ * shmatch.c -- shell interface to posix regular expression matching.
+ */
+
 /* Copyright (C) 2003 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
-   Bash is free software; you can redistribute it and/or modify it under
-   the terms of the GNU General Public License as published by the Free
-   Software Foundation; either version 2, or (at your option) any later
-   version.
+   Bash is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
 
-   Bash is distributed in the hope that it will be useful, but WITHOUT ANY
-   WARRANTY; without even the implied warranty of MERCHANTABILITY or
-   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
-   for more details.
+   Bash is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License along
-   with Bash; see the file COPYING.  If not, write to the Free Software
-   Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */
-
-/*
- * shmatch.c -- shell interface to posix regular expression matching.
- */
+   You should have received a copy of the GNU General Public License
+   along with Bash.  If not, see <http://www.gnu.org/licenses/>.
+*/
 
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
@@ -39,7 +39,7 @@
 #include "variables.h"
 #include "externs.h"
 
-extern int glob_ignore_case;
+extern int glob_ignore_case, match_ignore_case;
 
 int
 sh_regmatch (string, pattern, flags)
@@ -57,15 +57,14 @@ sh_regmatch (string, pattern, flags)
   char *subexp_str;
   int subexp_len;
 #endif
-  int result;  
-
+  int result;
 
 #if defined (ARRAY_VARS)
   rematch = (SHELL_VAR *)NULL;
 #endif
-  
+
   rflags = REG_EXTENDED;
-  if (glob_ignore_case)
+  if (glob_ignore_case || match_ignore_case)
     rflags |= REG_ICASE;
 #if !defined (ARRAY_VARS)
   rflags |= REG_NOSUB;
